import type { Ref, ComputedRef } from 'vue'
import type { LightSource } from '../types.d'
import { BLOCK_SIZE, STAGE_WIDTH, STAGE_HEIGHT } from '../level/def'

type RefOrComputed<T> = Ref<T> | ComputedRef<T>

export default function useLightMask(
  ctx: CanvasRenderingContext2D,
  y: RefOrComputed<number>,
  tx: RefOrComputed<number>,
  ty: RefOrComputed<number>,
  lightBarrier: RefOrComputed<number[]>,
  lightSources: RefOrComputed<LightSource[]>,
) {
  const W = ((STAGE_WIDTH + 2) * BLOCK_SIZE)
  const H = ((STAGE_HEIGHT + 2) * BLOCK_SIZE)
  const B = BLOCK_SIZE - 4 // no idea why there is a difference, but it is 4px
  const BHalf = B / 2

  const playerX = (W - B) / 2 + B / 4
  const playerY = H / 2 - BHalf
  const playerLightSize = B * 1.8

  function drawPlayerLight(sizeMul:number) {
    const playerLight = ctx.createRadialGradient(
      playerX - tx.value, playerY - ty.value, 0,
      playerX - tx.value, playerY - ty.value, playerLightSize * sizeMul
    )

    playerLight.addColorStop(0.7, "#FFFA");
    playerLight.addColorStop(1, "#FFF0");

    // Set the fill style and draw a rectangle
    ctx.fillStyle = playerLight;
    ctx.fillRect(0, 0, W, H)
  }

  function drawLightSources() {
    for (const src of lightSources.value) {
      const x = src.x * B
      const y = src.y * B
      const strength = src.strength + (src.strength * Math.random()) / 10
      const light = ctx.createRadialGradient(
        x + BHalf, y - BHalf, 0,
        x + BHalf, y - BHalf, strength * B,
      )
      const color = src.color

      light.addColorStop(0.0, color)
      light.addColorStop(0.4, `${color}A`)
      light.addColorStop(1, `${color}0`)

      ctx.fillStyle = light
      ctx.fillRect(0, 0, W, H)
    }
  }

  function drawShadows() {
    const barrier = lightBarrier.value

    for (let col = 0; col < barrier.length; col++) {
      const level = (barrier[col] - y.value) * B
      const x = B*col

      // gradient for the shadow that is cast down from the surface
      const gradient = ctx.createLinearGradient(0, 0, 0, H)
      gradient.addColorStop(0, '#FFF')
      gradient.addColorStop(Math.min(level / H, 1), '#FFF')
      gradient.addColorStop(Math.min((level + B) / H, 1), '#000')
      ctx.fillStyle = gradient
      ctx.fillRect(x, 0, B, H)
    }
  }

  return function update() {
    // first, throw the world in complete darkness
    ctx.fillStyle = '#FFFFFF'
    ctx.fillRect(0, 0, W, H)

    // second, hide what is beneath
    drawShadows()

    // third, fight the darkness
    drawLightSources()

    // finally, draw the players light
    // with a size multiplicator which might be later used to
    // simulate greater illumination with candles or torches
    drawPlayerLight(1)
  }
}
